
/**
 *
 * A jukebox program
 *
 * Protocall 0-0-2 : added confirmations
 *
 **/

import java.util.*;
import java.net.*;
import java.io.*;
import javazoom.jl.player.advanced.*;
import javazoom.jl.player.*;
import javazoom.jl.decoder.*;


/* usage: instantiate, listenForConnections, startPlaying */

public class PlayMusicD
{
      String musicroot;
      String tagdir;

      LinkedList toplay;
      LinkedList history;

      private boolean toplayAvail = true;
      private boolean historyAvail = true;

      Vector extentions;

      boolean done = false;
      int debug;

      Player player;

      Server server;

      Collection allSongs;

      public static int DEFAULT_PORTNUM = 10332;

      public PlayMusicD(String musicroot, int debug, String extentions, String tagdir)
      {
	 this.musicroot = musicroot;
	 this.tagdir = tagdir;
	 this.debug = debug;

	 this.extentions = new Vector();
	 StringTokenizer st = new StringTokenizer(extentions);
	 while (st.hasMoreTokens())
	    this.extentions.add((String)st.nextToken());

	 toplay = new LinkedList();
	 history = new LinkedList();

	 updateAllSongs();
      }

      public void listenForConnections(int portNum, String allow, String deny)
      {
	 server = new Server(portNum, allow, deny,this);
	 server.start();
      }

      public void startPlaying()
      {
	 player = new Player(toplay,history,this);
	 player.start();
      }

      public synchronized boolean requestToplay()
      {
	 if(toplayAvail)
	 {
	    toplayAvail = false;
	    return true;
	 }
	 return false;	 
      }
      public synchronized boolean requestHistory()
      {
	 if(historyAvail)
	 {
	    historyAvail = false;
	    return true;
	 }
	 return false;	 
      }

      public void doneWithToplay()
      {
	 toplayAvail = true;
      }
      public void doneWithHistory()
      {
	 historyAvail = true;
      }

      

      public void tagsUpdated()
      {
	 server.broadcast("Update tags");
      }
      public void queueUpdated()
      {
	 server.broadcast("Update queue");
      }     

      public String pause()
      {
	 String retval = player.pause();
	 if(retval == null)
	    return "1 Player paused";
	 else
	    return "0 " + retval;
      }

      public String resume()
      {
	 String r = player.unpause();
	 if(r == null)
	    return "1 Player unpaused";
	 else
	    return "0 " + r;
      }

      public String getNowPlaying()
      {
	 String r = player.getNowPlaying();
	 if(r != null)
	    return r + "\n1 NowPlaying sent";
	 return "0 Can't get now playing";	 
      }
      
      public String clearQueue()
      {
	 if(toplay != null)
	 {
	    while(!requestToplay())
	       Utility.sleep(50);
	    toplay.clear();
	    doneWithToplay();

	    queueUpdated();
	    return "1 Queue cleared";
	 }
	 return "0 Can't clear queue";
      }

      public String clearHistory()
      {
	 if(history != null)
	 {
	    while(!requestHistory())
	       Utility.sleep(50);
	    history.clear();
	    doneWithHistory();

	    queueUpdated();
	    return "1 History cleared";
	 }
	 return "0 Can't clear history";
      }

      public String removeFromToplay(String strIdx)
      {
	 if(toplay == null)
	    Utility.die("Toplay shouldn't be null in removeFromToplay");

	 int idx = -1;
	 try
	 {
	    idx = Integer.parseInt(strIdx);
	 }
	 catch(NumberFormatException n)
	 {
	    return "0 Not a number";
	 }
	 if( idx < 0 || idx >= toplay.size() )
	    return "0 toplay element out of range";

	 while(!requestToplay())
	    Utility.sleep(50);
	 if( idx < 0 || idx >= toplay.size() )
	    return "0 toplay element out of range";
	 toplay.remove(idx);
	 doneWithToplay();

	 return "1 Song removed"; 
      }

      public String removeFromHistory(String strIdx)
      {
	 if(history == null)
	    Utility.die("history shouldn't be null in removeFromHistory");

	 int idx = -1;
	 try
	 {
	    idx = Integer.parseInt(strIdx);
	 }
	 catch(NumberFormatException n)
	 {
	    return "0 Not a number";
	 }
	 if( idx < 0 || idx >= history.size() )
	    return "0 history element out of range";

	 while(!requestHistory()) 
	    Utility.sleep(50); 
	 if( idx < 0 || idx >= history.size() )
	    return "0 history element out of range";
	 history.remove(idx);
	 doneWithHistory();

	 return "1 Song removed";
      }

      public String queueUpSong(String s)
      {
	 if(toplay == null)
	    Utility.die("toplay shouldn't be null in queueUpSong");

	 if (s.indexOf("..") == -1)
	 {
	    boolean wasEmpty = toplay.isEmpty();

	    if(! (new File(musicroot + s)).exists())
	       return "0 File not found";
	    
	    while(!requestToplay()) 
	       Utility.sleep(50); 
	    toplay.addLast(musicroot + s);
	    doneWithToplay(); 

	    queueUpdated();

	    if(wasEmpty && player.nowPlaying == null)
	    {
	       if(debug > 1)
		  System.out.println("Going to go set up the next song");
	       String r = player.setUpNextSong();
	       if(r == null)
		  return "1 Song starting";
	       else
		  return "0 " + r;
	    }
	    return "1 Song queued";
	 }
	 else
	    return "0 Not in music subdir";
      }

      public String nextSong()
      {
	 String r = player.nextSong();
	 if(r == null)
	    return "1 Advanced to next song";
	 return "0 " + r;	 
      }

      public String nextSongN(String s)
      {
	 int n = -1;
	 try
	 {
	    n = Integer.parseInt(s);
	 }
	 catch(NumberFormatException nfe)
	 {
	    return "0 Not a number";
	 }
	 String	 r = player.nextSongN(n);
	 if(r == null)
	    return "1 Advanced to " + n + (n == 1 ? "th" : (n == 2 ? "nd" : (n == 3 ? "rd" : "th"))) + " song";
	 return r;
      }

      public String previousSong()
      {
	 String r = player.previousSong();
	 if(r == null)
	    return "1 Reversed to previous song";
	 return r;
      }

      public String previousSongN(String s)
      {
	 int n = -1;
	 try
	 {
	    n = Integer.parseInt(s);
	 }
	 catch(NumberFormatException nfe)
	 {
	    return "0 Not a number";
	 }
	 String	 r = player.previousSongN(n);
	 if(r == null)
	    return "1 Reversed to " + n + (n == 1 ? "th" : (n == 2 ? "nd" : (n == 3 ? "rd" : "th"))) + " song";
	 return r;
      }


      public void updateAllSongs()
      {
	 allSongs = new TreeSet(getAllSongs(new File (musicroot))); 
      }

      public String getStatus()
      {
	 return player.getStatus() + "\n1 Status sent";
      }


      public String refreshSongList()
      {
	 updateAllSongs();
	 return "1 Song list re-read";
      }

      public String listSongs()
      {
	 StringBuffer tr = new StringBuffer();

	 if(allSongs == null)
	    return "0 Can't list music files";

	 Iterator it = allSongs.iterator();
	 while(it.hasNext())
	    tr.append( ((String)it.next()).trim() + "\n" );

	 tr.append("1 Songs listed");
         return tr.toString();
      }


      private Collection getAllSongs(File f)
      {
	 return getAllSongsH(f, new HashSet());
      }
      private Collection getAllSongsH(File f, Collection sofar)
      {
	 if (f.isDirectory())
	 {
	    File[] fs = f.listFiles();
	    if(fs == null)
	       return null;
	    for (int i = 0 ; i < fs.length ; i++)
	       if(getAllSongsH(fs[i],sofar) == null)
		  return null;
	 }
	 else
	 {
	    for ( int i = 0 ; i < extentions.size() ; i++)
	       if (f.getName().endsWith((String)extentions.get(i)))
	       {
		  if (f.getPath().startsWith(musicroot))
		     sofar.add(f.getPath().substring(musicroot.length(), f.getPath().length()));
		  else if(debug > 1)
		     System.out.println("A file out of range snuck in: " + f.getPath());
	       }
	 }
	 return sofar;
      }


      private Vector getTagFiles()
      {
	 Vector tf = new Vector();
	 File[] tagFiles = new File(tagdir).listFiles();
	 
	 if(tagFiles == null)
	    return null;
	 
	 for(int i = 0 ; i < tagFiles.length ; i++)
	    if(tagFiles[i].getName().endsWith(".tgf") && !tagFiles[i].isDirectory())
	       tf.add(tagFiles[i].getName().substring(0,tagFiles[i].getName().length() - (".tgf".length())));
	 
	 return tf;
      }

      private Collection getSongsForTag(String tag)
      {
	 tag = tag.toLowerCase();

	 File tagFile = new File(tagdir, tag + ".tgf");
	 Collection sngs = new HashSet();
	 
	 try
	 {
	    BufferedReader br = new BufferedReader(
	       new InputStreamReader(
		  new FileInputStream(tagFile)));

	    String cline = null;
	    while( (cline=br.readLine()) != null)
	       sngs.add(cline);
	    br.close();
	 }
	 catch (FileNotFoundException e)
	 {
//	    if(debug > 0)
//	       System.out.println("No songs for tag " + tag);
	 }
	 catch (IOException e)
	 {
	    if(debug > 0)
	       System.out.println("Error reading tag file " + tag);
	    return null;
	 }

	 return sngs;

      }

      public String listTags()
      {
	 StringBuffer tr = new StringBuffer();

	 Collection tf = getTagFiles();
	 
	 if(tr == null)
	    return "0 Can't list tag files";

	 Iterator it = (new TreeSet(tf)).iterator();

	 while(it.hasNext())
	    tr.append(it.next() + "\n");	 

	 tr.append("1 Tag files listed");

	 return tr.toString();	       
      }

      public String listTagsForSong(String song)
      {
	 StringBuffer tags = new StringBuffer();
	 Collection tf = getTagFiles();

	 TreeSet tr = new TreeSet();

	 if(tf == null)
	    return "0 Can't list tag files";
	 
	 Iterator trIt = tf.iterator();
	 while(trIt.hasNext())
	 {
	    String c = (String)trIt.next();
	    Collection sft = getSongsForTag(c);
	    
	    if(debug>1)
	       System.out.println(c + " : " + sft);

	    if(sft == null)
	       return "0 Can't read tag files";

	    if(sft.contains(song))
	    {
	       if(debug > 1)
		  System.out.println("Decided the " + c + " tag applied to " + song);
	       tr.add(c);
	    }
	 }
	 
	 Iterator it = tr.iterator();
	 while(it.hasNext())
	    tags.append(it.next() + "\n");	 

	 tags.append("1 Tags listed");

	 return tags.toString();
      }

      public String deleteTag(String tag)
      {
	 tag = tag.toLowerCase();

	 File tagfile = new File(tagdir,tag + ".tgf");
	 if(tagfile.exists())
	 {
	    File newfile = new File(tagdir,tag + ".tgf.old");
	    if(!tagfile.renameTo(newfile))
	    {
	       if(debug > 0)
		  System.out.println("Couldn't delete/rename the tag file " + tagfile +  " to " + newfile );
	       return "0 Tag couldn't be deleted";
	    }
	    else
	    {
	       tagsUpdated();
	       return "1 Tag deleted";
	    }
	 }
	 else 
	    return "0 Tag does not exist";
      }

      public String newTag(String tag)
      {
	 tag = tag.toLowerCase();

	 File tagfile = new File(tagdir,tag + ".tgf");

	 try
	 {
	    if(!tagfile.exists())
	       tagfile.createNewFile();
	    tagsUpdated();
	    return "1 Tag created";
	 }
	 catch (IOException ioe)
	 {
	    if(debug > 0)
	       System.out.println("Couldn't create tag file " + tagfile + " Are permissions right?");
	    return "0 Can't create tag file";
	 }
      }

      private String parseTag(String tagAndSong)
      {
	 if(tagAndSong.indexOf(' ') == -1)
	 {
	    if(debug>0)
	       System.out.println("parseTag: given invalid value \"" + tagAndSong + "\"");
	    return null;
	 }

	 return tagAndSong.substring(0,tagAndSong.indexOf(' '));	 
      }

      private String parseSong(String tagAndSong)
      {
	 if(tagAndSong.indexOf(' ') == -1)
	 {
	    if(debug>0)
	       System.out.println("parseTag: given invalid value \"" + tagAndSong + "\"");
	    return null;
	 }

	 return tagAndSong.substring(tagAndSong.indexOf(' ') + 1, tagAndSong.length());
      }

      public String tag(String tagAndSong)
      {

	 String tag = parseTag(tagAndSong);
	 if(tag == null)
	    return "0 Can't parse tag";

	 String song = parseSong(tagAndSong);
	 if(song == null)
	    return "0 Can't parse song";

	 if(tag == "" || song == "")
	    return "0 Tag and song not understood";

	 tag = tag.toLowerCase();
	 
	 Collection sft = getSongsForTag(tag);
	 if(sft == null)
	    return "Can't read tagfile";

	 if(sft.contains(song))
	 {
	    if(debug > 0)
	       System.out.println("tag: song " + song + " already tagged by " + tag );
	    return "0 This song is already tagged by this tag";
	 }

	 // append
	 File f = new File(tagdir, tag + ".tgf");
	 try
	 {
	    System.out.println(tagdir);
	    if(!f.exists())
	       f.createNewFile();
	    PrintWriter pw = new PrintWriter(
	       new OutputStreamWriter(
		  new FileOutputStream(f, true)));
	    
	    pw.println( song );
	    
	    pw.flush();
	    pw.close();
	 }
	 catch (FileNotFoundException e)
	 {
	    Utility.die("File not found, but we should have created it: " + f);
	 }
	 catch (IOException e)
	 {
	    if(debug > 0)
	       System.out.println("Couldn't append song \"" + song + "\" to taglist \"" + tag 
				  + "\".  Are you sure permissions are right?");
	    if(debug > 2)
	       e.printStackTrace();
	    return "0 Couldn't append song to taglist.";
	 }
	 tagsUpdated();
	 return "1 Song tagged";
      }

      public String untag(String tagAndSong)
      {
	 String tag = parseTag(tagAndSong); 
         String song = parseSong(tagAndSong); 
 
	 if(song == null) 
            return "0 Can't parse song"; 

	 if(tag == null) 
            return "0 Can't parse tag"; 


         if(tag == "" || song == "") 
            return "0 Tag and song not understood";

	 tag = tag.toLowerCase();

	 Collection gsft = getSongsForTag(tag);
	 if(gsft == null)
	    return "0 Can't read tag file";
	 
	 if(!gsft.contains(song)) 
         { 
            if(debug > 0) 
               System.out.println("untag: song " + song + " not already tagged by " + tag ); 
	    return "0 This song is not tagged by this tag";
         } 

	 //construct the replacement taglist
	 StringBuffer tr = new StringBuffer();
	 
	 Iterator it = gsft.iterator();
	 while(it.hasNext())
	 {
	    String cSong = (String)it.next();
	    if(!cSong.equals(song))
	       tr.append(cSong + "\n");
	 }
	 
	 // don't append ; replace

	 File f = new File(tagdir, tag + ".tgf");

	 if (f.exists())
	    f.delete();

	 try
	 {
	    f.createNewFile();
	 }
         catch (IOException e) 
         { 
            if(debug > 0) 
               System.out.println("Couldn't open song \"" + song + "\" to taglist \"" + tag  
                                  + "\".  Are you sure permissions are right?"); 
	    return "0 Couldn't overwrite tagfile with replacement";
         } 

	 try
	 {
	    PrintWriter pw = new PrintWriter(
	       new OutputStreamWriter(
		  new FileOutputStream(f, false)));
	 
	    
	    pw.println( tr.toString().trim() );
	    
	    pw.flush();
	    pw.close();
	 }
	 catch(FileNotFoundException floop)
	 {
	    Utility.die("File not found, but we should have created it: " + f); 
	 }

	 tagsUpdated();
	 return "1 Song untagged";
	 
      }

      /*
       * Format for expr is intuitive, but a bit complicated.
       *   Some examples:
       *     foo       
       *      - returns all songs containing foo in their filename
       *     t:bar
       *      - returns all songs tagged as bar
       *     foo && t:bar
       *      - returns all with foo which also are tagged as bar
       *     t:bar && !baz
       *      - returns all tagged as foo which do not contain baz
       *     (t:bar && !baz) || t:fleem
       *      - and it gets more and more messy
       *
       */

      public String tagSearch(String expr)
      {
	 StringBuffer tr = new StringBuffer();
	 Collection res = tagSearchH(expr);

	 if(res == null)
	    return "0 Search failed";

	 Iterator it = (new TreeSet(res)).iterator();

	 while(it.hasNext())
	    tr.append((String)it.next() + "\n");

	 tr.append("1 Search completed");

	 return (tr.toString());
      }

      private Collection getSongsForKeyword(String word)
      {
	 TreeSet sofar = new TreeSet();

	 if(allSongs == null)
	    return null;

	 Iterator it = allSongs.iterator();
	 while(it.hasNext())
	 {
	    String curSong = (String)it.next();
	    if( curSong.toLowerCase().indexOf(word.toLowerCase()) != -1)
	       sofar.add(curSong);
	 }
	 return sofar;
      }

      private boolean isAtomic(String expr)
      {
	 return expr.indexOf("&&") == -1 && expr.indexOf("||") == -1 && expr.indexOf('!') == -1
	    && expr.indexOf('(') == -1 && expr.indexOf(')') == -1;
      }

      private Collection tagSearchH(String expr)
      {
	 if(expr.length() == 0)
	    return new HashSet();

	 expr = removeOuterParens(expr).trim();
	 
	 if( isAtomic(expr) )
	 {
	    if(expr.startsWith("t:"))
	       return getSongsForTag(expr.substring("t:".length(),expr.length()));
	    else
	       return getSongsForKeyword(expr);
	 }

	 if(expr.charAt(0) == '!' && expr.length() > 1 &&
	    (hasMatchedOuterParens(expr.substring(1,expr.length())) || isAtomic(expr.substring(1,expr.length())) ) )
	 {
	    if(allSongs == null)
	       return null;
	    
	    Collection subset = new HashSet(allSongs);

	    subset.removeAll(tagSearchH(expr.substring("!".length(),expr.length())));
	    return subset;
	 }

	 // go until you get to an operator when there are no open parens,
	 // then return (op (eval "before op") (eval "after op"))

	 int actOpInd = findActiveOperator(expr);
	 if(actOpInd == -1)
	    return null;
	 char operator = expr.charAt(actOpInd);
	 
	 String before = expr.substring(0,actOpInd).trim();

	 if(actOpInd+2 >= expr.length())
	    return null;

	 String after  = expr.substring(actOpInd+2,expr.length()).trim();

	 Collection evaldBefore = tagSearchH(before);
	 Collection evaldAfter  = tagSearchH(after);

	 if(evaldBefore == null || evaldAfter == null)
	    return null;
	 
	 if(operator == '&')
	    evaldBefore.retainAll(evaldAfter);
	 else // operator == '|'
	    evaldBefore.addAll(evaldAfter);

	 return evaldBefore;
      }

      private int findActiveOperator(String expr)
      {
	 int parens = 0;
	 int prevChar = -1;
	 for (int i = 0 ; i < expr.length() ; i++)
	 {
	    if(expr.charAt(i) == '(')
	       parens++;
	    else if(expr.charAt(i) == ')')
	       parens--;
	    else if(  ((expr.charAt(i) == '&' && prevChar == '&') || (expr.charAt(i) == '|' && prevChar == '|') ) 
		      && parens == 0)
	       return i-1;
	    prevChar = expr.charAt(i);
	 }
	 return -1;
      }

      private boolean hasMatchedOuterParens(String expr)
      {
	 if(!(expr.startsWith("(") && expr.endsWith(")")))
	    return false;

	 int parens = 0;
	 for(int i = 0 ; i < expr.length() ; i++)
	 {
	    if (expr.charAt(i) == '(')
	       parens++;
	    else if(expr.charAt(i) == ')')
	    {
	       parens--;
	       if(parens == 0 && i != expr.length() - 1)
		  return false;
	    }
	 }
	 return true;

      }

      private String removeOuterParens(String expr)
      {
	 if(hasMatchedOuterParens(expr))
	    return expr.substring(1, expr.length()-1);
	 return expr;
      }

      public String getQueue()
      {
	 return getList(toplay, "Queue");
      }

      public String getHistory()
      {
	 return getList(history, "History");
      }

      public String stopp()
      {
	 return player.stopp();
      }

      public String startt()
      {
	 return player.startt();
      }

      public String getList(Collection c, String collType)
      {
	 Iterator i = c.iterator();
	 StringBuffer tr = new StringBuffer();

	 while (i.hasNext())
	 {
	    String nxt = (String)i.next();
	    if(nxt.startsWith(musicroot))
	    {
	       tr.append(nxt.substring(musicroot.length(),nxt.length()));
	       tr.append("\n");
	    }
	    else if(debug > 1) 
	       System.out.println("0 A file out of range snuck in: " + nxt); 
	 }
	 
	 tr.append("1 " + collType + " listed");
	 return tr.toString();
      }

      public static boolean validateIPs(String s)
      {

	 StringTokenizer st = new StringTokenizer(s," ");

	 while (st.hasMoreTokens())
	 {
	    if(!validateIP(st.nextToken())) 
	       return false;
	 }
	 
	 return true;
      }

      public static boolean validateIP(String ip)
      {
	 if ("ALL".equals(ip))
	    return true;
	 if ("NONE".equals(ip))
	    return true;

	 int numDots = 0;
	 for(int i = 0 ; i < ip.length() ; i++)
	    if (ip.charAt(i) == '.')
	       numDots++;
	 if (numDots != 3)
	    return false;
	 else
	    return true;
      }


      public static void main(String[] args)
      {
	 Hashtable config = null;

	 if (args.length > 0)
	 {
	    if (args[0].equals("--help"))
	    {
	       System.out.println("PlayMusicD : A program to remotely control a music player");
	       System.out.println("  If a file is passed in on the command line, it is used as");
	       System.out.println("  the configuration file.  Otherwise /etc/playmusicd.conf is used.");
	       System.exit(0);
	    }
	    else if(args[0].equals("--stop"))
	    {
	       try
	       {
		  int port = 10332;
		  if(args.length == 2)
		     port = Integer.parseInt(args[1]);
		  
		  Socket s = new Socket();
		  s.connect(new InetSocketAddress("localhost",port));
		  BufferedReader in = new BufferedReader( new InputStreamReader( s.getInputStream()));
		  PrintWriter out = new PrintWriter( new OutputStreamWriter(s.getOutputStream()), true);
		  
		  String ret = in.readLine();
		  if(ret == null || !ret.startsWith("1 "))
		  {
		     System.err.println("Couldn't connect: " + ret);
		     System.exit(3);
		  }
		  
		  out.println("KillDeamon");
		  
		  ret = in.readLine();
		  if(ret == null || ret.startsWith("0 "))
		  {
		     System.err.println("Couldn't kill daemon: " + ret);
		     System.exit(4);
		  }
		  else if(ret.startsWith("1 "))
		  {
		     System.out.println("Deamon killed");
		     System.exit(0);
		  }
		  else
		  {
		     System.err.println("Protocoll not understood: " + ret);
		     System.exit(5);
		  }
	       }
	       catch(NumberFormatException nfe)
	       {
		  System.err.println("Port not understood as number : " + args[1]);
		  System.exit(6);
	       }
	       catch(IOException ioe)
	       {
		  System.err.println("Communication error.  Sorry");
		  System.exit(7);
	       }
	    }
	    else 
	    {
	       // turns keys to lowercase (among other things)
	       config = Utility.parseConfig(args[0]);

	    }
	 }
	 else if( (new File("/etc/playmusicd.conf")).exists() )
	    config = Utility.parseConfig("/etc/playmusicd.conf");
	 else
	 {					
	    System.out.println("No configuration file found.  Please put one in /etc/playmusicd.conf");
	    System.out.println("or pass the location of one in on the commandline.");
	    System.exit(2);
	 }

	 if (config == null)
	    Utility.die("Config is null after reading and shouldn't be");

	 int debug = Integer.parseInt(Utility.configSet(config,"debug","2"));
	 if (debug > 4 || debug < 0)
	    Utility.die("Invalid debug value : " + debug);

	 int port = Integer.parseInt(Utility.configSet(config,"port","10332"));
	 if (port > 65535 || port < 1)
	    Utility.die("Invalid port value : " + port);

	 String allow = Utility.configSet(config, "allow", "127.0.0.1");
	 if(!validateIPs(allow))
	    Utility.die("Invalid allow value : " + allow);

	 String deny = Utility.configSet(config, "deny", "ALL");
	 if(!validateIPs(deny))
	    Utility.die("Invalid deny value : " + deny);

	 String musicroot = Utility.configSet(config, "musicroot", "/mnt/music");
	 if( !(new File(musicroot)).exists() )
	    Utility.die("musicroot does not exist : " + musicroot);

	 String tagdir = Utility.configSet(config, "tagdir", "/var/playmusicd/tagdir");
	 if( !(new File(tagdir)).exists() )
	    Utility.die("tagdir does not exist : " + musicroot);
	 if( !(new File(tagdir)).isDirectory() )
	    Utility.die("tagdir is a file ; should be a directory : " + musicroot);

	 String extentions = Utility.configSet(config, "extentions", "mp3");
	 if (!"mp3".equals(extentions))
	    Utility.die("Only mp3 is supported so far : " + extentions);
	 

	 if(debug > 1)
	 {
	    System.out.println("  debug       = " + debug);
	    System.out.println("  port        = " + port);
	    System.out.println("  allow       = " + allow);
	    System.out.println("  deny        = " + deny);
	    System.out.println("  musicroot   = " + musicroot);
	    System.out.println("  tagdir      = " + tagdir);
    	    System.out.println("  extentions  = " + extentions);
	 }
	    
	 PlayMusicD p = new PlayMusicD(musicroot,debug,extentions,tagdir);
	 
	 p.listenForConnections(port, allow, deny);

	 p.startPlaying();
      }
}

class ConnectionHandler extends Thread
{
      Socket tsock;
      PlayMusicD deamon;
      boolean connectionValidated = false;
      boolean done = false;

      long lastCommand = System.currentTimeMillis();

      BufferedReader br;
      PrintWriter pw;

      LinkedList notifications = new LinkedList();

      String helpMessage = 
	 "This is a music player deamon.  Valid commands are:\n" +
	 "  Pause\n" +
	 "  Resume\n" + 
         "  Stop\n" +
         "  Start\n" +
	 "  ClearQueue\n" +
	 "  QueueUpSong <song>\n" + 
	 "  Status\n" + 
         "  RefreshSongList\n" +
	 "  ListSongs\n" + 
	 "  ListTags\n" +
	 "  ListTagsForSong <song>\n" +
	 "  Tag <tag> <song>\n" +
	 "  Untag <tag> <song>\n" +
	 "  TagSearch <expr>\n" +
	 "  GetQueue\n" + 
	 "  KillDeamon\n" +
	 "  CheckNotifications\n" + 
	 "  KeepAlive\n" + 
	 "  RemoveFromHistory <idx>\n" +
	 "  RemoveFromToplay <idx>\n" +
	 "  Unmute\n" + 
	 "  PreviousSong\n" + 
	 "  NextSong\n" + 
	 "  PreviousSong <jmp>\n" + 
	 "  NextSong <jmp>\n\n";

      public ConnectionHandler(PlayMusicD deamon, Socket tsock, String allow, String deny)
      {
	 this.tsock = tsock;
	 this.deamon = deamon;

	 connectionValidated = allowConnection(tsock.getInetAddress().toString(), allow, deny);

	 try
	 {
	    br = new BufferedReader(new InputStreamReader(tsock.getInputStream()));
	    
	    //the boolean is to turn on autoflush
            pw = new PrintWriter(tsock.getOutputStream(),true);
	 }
	 catch(IOException ioe)
	 {
	    System.err.println("Couldn't make buffered readers and writers : " + tsock.getInetAddress());
	    done = true;
	 }
      }

      public void tellDead()
      {
	 done = true;
      }

      public boolean hasExpired()
      {
	 // A five minute timeout
	 return System.currentTimeMillis() - lastCommand > 5 * 60 * 1000;	    
      }

      public boolean allowConnection(String addr, String allow, String deny)
      {

	 if (addr == null)
	 {
	    if(deamon.debug > 0)
	       System.out.println("Recieved invalid null IP");
	    return false;
	 }

	 while(addr.startsWith("/"))
	    addr = addr.substring(1, addr.length());

	 // allow connections if they are explicity allowed or not explicitly denied.
	 return isMatched(addr, allow) || (! isMatched(addr, deny));
      }

      public boolean isMatched (String addr, String filter)
      {
	 if (filter == null )
	    return false;

	 if(filter.equals(addr))
	    return true;

	 if(filter.equals("ALL"))
	    return true;

	 if(filter.equals("NONE"))
	    return false;

	 StringTokenizer st = new StringTokenizer(filter);
	 String curTok;
	 while(st.hasMoreTokens())
	 {
	    curTok = st.nextToken();
	    if (addr.equals(curTok))
	       return true;

	    if (curTok.indexOf("*") != -1)
	    {
	       StringTokenizer A = new StringTokenizer(addr,".");
	       StringTokenizer B = new StringTokenizer(curTok,".");
	       boolean matchedSoFar = true;
		  
	       String tmpTokA;
	       String tmpTokB;
	       for (int i = 0 ; i < 4 ; i++)
	       {
		  tmpTokA = A.nextToken();
		  tmpTokB = B.nextToken();
		  
		  matchedSoFar = matchedSoFar && tmpTokB.equals("*") ? true : tmpTokA.equals(tmpTokB);
	       }
	       if (matchedSoFar)
		  return true;
	    }
	 }
	    
	 return false;
      }

      public void notify(String s)
      {
	 notifications.addFirst(s);
      }

      public void run()
      {
	 if(!connectionValidated)
	 {
	    if (deamon.debug > 1)
	       System.out.println("Rejected connection from " + tsock.getInetAddress() );
	    tell("0 Connection rejected");
	    deamon.server.removeConnection(this);
	    return;
	 }
	 else
	    tell("1 Connection accepted");

	 while (! done)
	 {
	    try
	    {
	       while(!br.ready())
		  Utility.sleep(30);
	    }
	    catch(IOException ioe)
	    {
	       System.err.println("Can't check if the stream from " + tsock.getInetAddress() + " is ready\n");
	       done = true;
	    }

	    String recd = null;
	    try 
	    {
	       recd = br.readLine();
	    }
	    catch(IOException ioe)
	    {
	       System.err.println("Can't read from " + tsock.getInetAddress() + "\n");
	       done = true;
	    }

	    if(deamon.debug > 1 && recd != null)
	       System.out.println("  Recd: " + recd);


	    if(recd != null)
	    {
	       lastCommand = System.currentTimeMillis();
	       
	       if (recd.equals("Pause"))
		  tell(deamon.pause());
	       else if (recd.equals("Resume"))
		  tell(deamon.resume());

	       else if (recd.equals("ClearQueue"))
		  tell(deamon.clearQueue());
	       else if (recd.equals("ClearHistory"))
		  tell(deamon.clearHistory());
	       else if (recd.startsWith("RemoveFromHistory "))
		  tell(deamon.removeFromHistory(recd.substring("RemoveFromHistory ".length() ,recd.length())));
	       else if (recd.startsWith("RemoveFromToplay "))
		  tell(deamon.removeFromToplay(recd.substring("RemoveFromToplay ".length() ,recd.length())));

	       else if (recd.startsWith("QueueUpSong"))
		  tell(deamon.queueUpSong(recd.substring("QueueUpSong ".length(),recd.length())));
//	       else if (recd.startsWith("QueueUpPlaylist"))
//		  tell(deamon.queueUpPlaylist(recd.substring("QueueUpPlaylist ".length(),recd.length())));

	       else if (recd.equals("Status"))
		  tell(deamon.getStatus());

	       else if (recd.equals("RefreshSongList"))
		  tell(deamon.refreshSongList());

	       else if (recd.equals("ListSongs"))
		  tell(deamon.listSongs());
	       else if (recd.equals("ListPlaylists"))
		  tell(deamon.listTags());


	       else if(recd.equals("ListTags"))
		  tell(deamon.listTags());
	       else if(recd.startsWith("ListTagsForSong "))
		  tell(deamon.listTagsForSong(recd.substring("ListTagsForSong ".length(),recd.length())) );
	       else if(recd.startsWith("Tag "))
		  tell(deamon.tag(recd.substring("Tag ".length(),recd.length())));
	       else if(recd.startsWith("Untag "))
		  tell(deamon.untag(recd.substring("Untag ".length(),recd.length())));
	       else if(recd.startsWith("NewTag "))
		  tell(deamon.newTag(recd.substring("NewTag ".length(),recd.length())));
	       else if(recd.startsWith("DeleteTag "))
		  tell(deamon.deleteTag(recd.substring("DeleteTag ".length(),recd.length())));
	       else if(recd.startsWith("TagSearch "))
		  tell(deamon.tagSearch(recd.substring("TagSearch ".length(),recd.length())) );
	    

	       else if (recd.equals("GetQueue"))
		  tell(deamon.getQueue());
	       else if (recd.equals("GetHistory"))
		  tell(deamon.getHistory());
	       else if (recd.equals("GetNowPlaying"))
		  tell(deamon.getNowPlaying());

	       else if (recd.equals("NextSong"))
		  tell(deamon.nextSong());
	       else if (recd.equals("PreviousSong"))
		  tell(deamon.previousSong());

	       else if(recd.equals("Stop"))
		  tell(deamon.stopp());
	       else if(recd.equals("Start"))
		  tell(deamon.startt());

	       else if (recd.startsWith("NextSong "))
		  tell(deamon.nextSongN(recd.substring("NextSong ".length(),recd.length())));
	       else if (recd.startsWith("PreviousSong "))
		  tell(deamon.previousSongN(recd.substring("PreviousSong ".length(),recd.length())));

	       else if (recd.equals("KillDeamon"))
	       {
		  tell("1 I am slain.  You inconsiderate little...");
		  Utility.die("Killed by " + tsock.getInetAddress());
	       }
	       else if (recd.equals("KeepAlive"))
	       {
		  tell("1 Keeping alive");
	       } 
	       else if (recd.equals("CheckNotifications"))
	       {
		  if(notifications.isEmpty())
		     tell("1 No notifications");
		  else
		  {
		     // remove duplicates and sort
		     Iterator it = (new TreeSet(notifications)).iterator(); 
		     
		     while(it.hasNext())
			tell((String)it.next());
		     notifications.clear();

		     tell("1 Notifications sent");
		  }
	       }
	       else if(recd.equals("Help") || recd.equals("ListCommands"))
	       {
		  tell(helpMessage);
		  tell("1 You've been helped now.");
	       }
	       else 
		  tell("0 Not understood");
	    }
	 }

	 try
	 {
	    tsock.close();
	 }
	 catch(IOException ioe)
	 {
	    if (deamon.debug > 0)
	       System.err.println("Can't close connection with " + tsock.getInetAddress());
	 }
      }

      public void tell(String s)
      {
	 pw.println(s);

	 if(deamon.debug > 1)
	    System.out.println("  Sent: \n" + s);
      }    
}

class Player extends Thread
{
      public static final int PLAYING=1,STOPPED=2,PAUSED=3;
      
      LinkedList toplay;
      LinkedList history;

      boolean done = false;
      String nowPlaying = null;

      private int state = PLAYING;

      PlayMusicD deamon;

      TweakedPlayer aplayer = null;
   
      public Player(LinkedList toplay, LinkedList history, PlayMusicD deamon)
      {
	 this.toplay = toplay;
	 this.history = history;
	 this.deamon = deamon;
      }

      public void run()
      {
	 while (true)
	 {
	    if(aplayer == null && nowPlaying != null && !stopped())
	       startPlayingSong();
	    Utility.sleep(300);
	 }
      }

      public boolean stopped()
      {
	 return state == STOPPED;
      }

      public boolean paused()
      {
	 return state == PAUSED;
      }

      public boolean playing()
      {
	 return state == PLAYING;
      }


      public String setUpNextSong()
      {
	 return setUpNextSong(1);
      }

      public String setUpNextSong(int n)
      {	 
	 for (int i = 0 ; i < n ; i++)
	 {
	    if(nowPlaying != null)
	    {
               while(!deamon.requestHistory()) 
		  Utility.sleep(50); 
	       history.addLast(nowPlaying);
	       deamon.doneWithHistory(); 

	       nowPlaying = null;
	    }
	    
	    if(!toplay.isEmpty())
	    {
	       while(!deamon.requestToplay())
		  Utility.sleep(50);
	       if(!toplay.isEmpty())
		  nowPlaying = (String)toplay.removeFirst(); 
	       deamon.doneWithToplay();
	    }
	 }

	 nextPrevTerm();
	 return null;
      }
      
      private void nextPrevTerm()
      {
	 if(aplayer != null)
	 {
	    aplayer.close();
	    aplayer = null;
	 }
      }

      // returns null on sucess, error message otherwise
      public String setUpPreviousSong(int n)
      {
	 for(int i = 0 ; i < n ; i++)
	 {
	    if(nowPlaying != null)
	    {
	       while(!deamon.requestToplay()) 
                  Utility.sleep(50); 
	       toplay.addFirst(nowPlaying);
               deamon.doneWithToplay(); 

	       nowPlaying = null;
	    }
	    
	    if(!history.isEmpty())
	    {
               while(!deamon.requestHistory())  
                  Utility.sleep(50);  
	       if(!history.isEmpty()) 
		  nowPlaying = (String)history.removeLast();
	       deamon.doneWithHistory();
	    }
	 }

	 if(nowPlaying == null)
	    setUpNextSong();
	 else
	    nextPrevTerm();

	 return null;
      }

      public String setUpPreviousSong()
      {
	 return setUpPreviousSong(1);
      }

      public void startPlayingSong()
      {
	 try
	 {
	    aplayer = new TweakedPlayer(new FileInputStream(nowPlaying),this);
	    aplayer.play();
	    deamon.queueUpdated();
	 }
	 catch(FileNotFoundException fnfe)
	 {
	    System.err.println("Someone slipped me a nonexistant file.  For shame.");
	 }
	 catch(javazoom.jl.decoder.JavaLayerException jle)
	 {
	    System.err.println("The media player doesn't seem to like your file");
	 }
      }

      public String pause()
      {
	 if(aplayer != null)
	 {
	    state = PAUSED;
	    aplayer.pause();
	    return null;
	 }
	 return "Player is not available";
      }

      public String unpause()
      {
	 if(aplayer != null)
	 {
	    state = PLAYING;
	    aplayer.resume();
	    return null;
	 }
	 return "Player is not available";
      }

      public String getNowPlaying()
      {
	 if(nowPlaying == null)
	    return "Nothing";

	 if(nowPlaying.startsWith(deamon.musicroot))
	    return nowPlaying.substring(deamon.musicroot.length(),nowPlaying.length());
	 else
	 {
	    if (deamon.debug > 1)
	       System.out.println("A file not in music root snuck in: " + nowPlaying);
	    return nowPlaying;
	 }
      }

      public String previousSong()
      {
	 return setUpPreviousSong();
      }
      
      public String previousSongN(int n)
      {
	 return setUpPreviousSong(n);
      }

      public String nextSong()
      {
	 return setUpNextSong();
      }

      public String nextSongN(int n)
      {
	 return setUpNextSong(n);
      }


      public String stopp()
      {
	 if(stopped())
	    return "0 Already stopped";
	 state = STOPPED;

	 if(aplayer != null)
         {
            aplayer.close();
            aplayer = null;
         }

	 return "1 Playing stopped";
      }

      public String startt()
      {
	 if(!stopped())
	    return "0 Already playing";
	 state = PLAYING;
	 return "1 Playing resumed";
      }

      public String getStatus()
      {
	 return (nowPlaying == null ? "No File" : 
		 (stopped() ? "Stopped" : 
		  ((paused()? "Paused : " : "Playing : ") + nowPlaying)));
      }
}


// Hacked from AdvancedPlayer
class TweakedPlayer extends AdvancedPlayer
{
      private boolean paused = false;
      Player contact;

      public TweakedPlayer(InputStream stream, Player contact) throws JavaLayerException
      {
	 super(stream);
	 this.contact = contact;
      }

      // note: does NOT send the appropriate playbackstarted event
      public void play() throws JavaLayerException
      {
	 boolean ret = true;
	 while (ret)
	 {
	    while(paused)
	       Utility.sleep(150);
	    ret = decodeFrame();
	 }
	 close();
	 contact.setUpNextSong();
      }

      public void pause()
      {
	 paused = true;
      }

      public void resume()
      {
	 paused = false;
      }
}

class ConnectionCleanerUpper extends Thread
{
      boolean done = false;
      Server s;
      public ConnectionCleanerUpper(Server s)
      {
	 this.s = s;
      }

      public void run()
      {
	 while (!done)
	 {
	    Utility.sleep(2*60*1000);


	    Iterator it = s.getConnections().iterator();
	    Vector toremove = new Vector();
	    while(it.hasNext())
	    {
	       ConnectionHandler ch = (ConnectionHandler)it.next();
	       if(ch.hasExpired())
		  toremove.add(ch);
	    }
	    it = toremove.iterator();
	    while(it.hasNext())
	       s.removeConnection((ConnectionHandler)it.next());
	 }
      }
}

class Server extends Thread
{
      ServerSocket listener = null;
      PlayMusicD deamon = null;
      String allow = null;
      String deny = null;
      boolean done = false;

      Vector connections = new Vector();

      public Server(int portNum, String allow, String deny, PlayMusicD deamon)
      {
	 this.deamon = deamon;
	 this.allow = allow;
	 this.deny = deny;

	 try
	 {
	    listener = new ServerSocket(portNum);
	 }
	 catch (IOException ioe)
	 {
	    Utility.die("Can't listen on port " + portNum);
	 }
	 if(deamon.debug > 1)
	    System.out.println("Listening for connections...");
      }

      public Collection getConnections()
      {
	 return connections;
      }

      public void removeConnection(ConnectionHandler ch)
      {
	 ch.tellDead();
	 connections.remove(ch);
	 if(deamon.debug > 1)
	    System.out.println("Removing connection");
      }

      public void broadcast(String s)
      {
	 for (int i = 0 ; i < connections.size() ; i++)
	    ((ConnectionHandler)connections.get(i)).notify(s);
      }

      public void run()
      {
	 while (!done)
	 {

	    Utility.sleep(1000);

	    Socket tsock = null;
	    try
	    {
	       // wait for someone to try to connect
	       tsock = listener.accept();
	    }
	    catch (IOException ioe)
	    {
	       Utility.die("Server can't accept connection.  This is bad.");
	    }
	    	    
	    if(deamon.debug > 1)
	       System.out.println("Got connection from " + tsock.getInetAddress());
	    
	    ConnectionHandler ch = new ConnectionHandler(deamon,tsock,allow,deny);
	    ch.start();
	    connections.add(ch);
	 }
      }
}